@@ -9,15 +9,19 @@ PrivBayes: Private Data Release via Bayesian Networks. (2017)
 
 import numpy as np
 import pandas as pd
+import random
+import dill
 from sklearn.metrics import mutual_info_score
 from collections import namedtuple
+from joblib import Parallel, delayed
+
 
 import synthesis.synthesizers.utils as utils
 from synthesis.synthesizers._base import BaseDPSynthesizer
 from thomas.core import BayesianNetwork
 from diffprivlib.mechanisms import Exponential
 
-APPair = namedtuple('AttributeParentPair', ['node', 'parents'])
+APPair = namedtuple('APPair', ['attribute', 'parents'])
 
 
 class PrivBayes(BaseDPSynthesizer):
@@ -34,12 +38,13 @@ class PrivBayes(BaseDPSynthesizer):
     """
 
     def __init__(self, epsilon=1.0, theta_usefulness=4, epsilon_split=0.3,
-                 score_function='R', network_init=None, verbose=True):
+                 score_function='R', network_init=None, n_cpus=None, verbose=True):
         super().__init__(epsilon=epsilon, verbose=verbose)
         self.theta_usefulness = theta_usefulness
         self.epsilon_split = epsilon_split  # also called Beta in paper
         self.score_function = score_function  # choose between 'R' and 'MI'
         self.network_init = network_init
+        self.n_cpus = n_cpus
 
     def fit(self, data):
         data = self._check_input_data(data)
@@ -100,7 +105,7 @@ class PrivBayes(BaseDPSynthesizer):
                 # empty set - domain size of node violates theta_usefulness
                 if len(max_parent_sets) == 0:
                     ap_pairs.append(APPair(node, parents=None))
-                # [empty set] - no parents found that meet domain size restrictions
+                # [empty set] - no parents found that meets domain size restrictions
                 elif len(max_parent_sets) == 1 and len(max_parent_sets[0]) == 0:
                     ap_pairs.append(APPair(node, parents=None))
                 else:
@@ -114,8 +119,8 @@ class PrivBayes(BaseDPSynthesizer):
             sampled_pair = self._exponential_mechanism(ap_pairs, scores)
 
             if self.verbose:
-                print("Selected attribute: '{}' - with parents: {}\n".format(sampled_pair.node, sampled_pair.parents))
-            nodes_selected.add(sampled_pair.node)
+                print("Selected attribute: '{}' - with parents: {}\n".format(sampled_pair.attribute, sampled_pair.parents))
+            nodes_selected.add(sampled_pair.attribute)
             self.network_.append(sampled_pair)
         if self.verbose:
             print("Learned Network Structure\n")
@@ -124,7 +129,7 @@ class PrivBayes(BaseDPSynthesizer):
     def _max_domain_size(self, data, node):
         """Computes the maximum domain size a node can have to satisfy theta-usefulness"""
         node_cardinality = utils.cardinality(data[node])
-        max_domain_size = self.n_records_fit_ * (1 - self.epsilon_split) * self.epsilon / \
+        max_domain_size = (self.n_records_fit_ * (1 - self.epsilon_split) * self.epsilon) / \
                           (2 * len(self.columns_) * self.theta_usefulness * node_cardinality)
         return max_domain_size
 
@@ -155,12 +160,12 @@ class PrivBayes(BaseDPSynthesizer):
         nodes = set(X.columns)
 
         if self.network_init is not None:
-            nodes_selected = set(n.node for n in self.network_init)
+            nodes_selected = set(n.attribute for n in self.network_init)
             # print("Pre-defined network init: {}".format(self.network_))
             for i, pair in enumerate(self.network_init):
                 if self.verbose:
                     print("{}/{} - init node {} - with parents: {}".format(i + 1, len(nodes),
-                                                                           pair.node, pair.parents))
+                                                                           pair.attribute, pair.parents))
             self.network_ = self.network_init.copy()
             return nodes, nodes_selected
 
@@ -169,7 +174,7 @@ class PrivBayes(BaseDPSynthesizer):
         nodes_selected = set()
 
         root = np.random.choice(tuple(nodes))
-        self.network_.append(APPair(node=root, parents=None))
+        self.network_.append(APPair(attribute=root, parents=None))
         nodes_selected.add(root)
         if self.verbose:
             print("1/{} - Root of network: {}\n".format(X.shape[1], root))
@@ -182,17 +187,25 @@ class PrivBayes(BaseDPSynthesizer):
 
     def _compute_scores(self, data, ap_pairs):
         """Compute score for all ap_pairs"""
-        scores = np.empty(len(ap_pairs))
-
-        for idx, pair in enumerate(ap_pairs):
-            if pair.parents is None:
-                scores[idx] = 0
-            elif self.score_function == 'R':
-                scores[idx] = self.r_score(data, pair.node, pair.parents)
-            elif self.score_function == 'MI':
-                scores[idx] = self.mi_score(data, pair.node, pair.parents)
+        if self.n_cpus:
+            scores = Parallel(n_jobs=self.n_cpus)(delayed(self.r_score)(data, pair.attribute, pair.parents) for pair in ap_pairs)
+        else:
+            scores = [self.r_score(data, pair.attribute, pair.parents) for pair in ap_pairs]
         return scores
 
+    # def _compute_scores(self, data, ap_pairs):
+    #     """Compute score for all ap_pairs"""
+    #     scores = np.empty(len(ap_pairs))
+    #
+    #     for idx, pair in enumerate(ap_pairs):
+    #         if pair.parents is None:
+    #             scores[idx] = 0
+    #         elif self.score_function == 'R':
+    #                 scores[idx] = self.r_score(data, pair.attribute, pair.parents)
+    #         elif self.score_function == 'MI':
+    #             scores[idx] = self.mi_score(data, pair.attribute, pair.parents)
+    #     return scores
+
     def _exponential_mechanism(self, ap_pairs, scores):
         """select APPair with exponential mechanism"""
         local_epsilon = self.epsilon * self.epsilon_split / self._n_nodes_dp_computed
@@ -208,17 +221,17 @@ class PrivBayes(BaseDPSynthesizer):
 
         for idx, pair in enumerate(self.network_):
             if pair.parents is None:
-                attributes = [pair.node]
+                attributes = [pair.attribute]
             else:
-                attributes = [*pair.parents, pair.node]
+                attributes = [*pair.parents, pair.attribute]
 
             cpt_size = utils.cardinality(data[attributes])
             if self.verbose:
                 print('Learning conditional probabilities: {} - with parents {} '
-                      '~ estimated size: {}'.format(pair.node, pair.parents, cpt_size))
+                      '~ estimated size: {}'.format(pair.attribute, pair.parents, cpt_size))
 
             dp_cpt = utils.dp_conditional_distribution(data[attributes], epsilon=local_epsilon)
-            self.cpt_[pair.node] = dp_cpt
+            self.cpt_[pair.attribute] = dp_cpt
         return self
 
     def _generate_data(self, n_records):
@@ -232,23 +245,25 @@ class PrivBayes(BaseDPSynthesizer):
             data_synth[i] = list(record.values())
 
         # numpy.array to pandas.DataFrame with original column ordering
-        data_synth = pd.DataFrame(data_synth, columns=[c.node for c in self.network_])
+        data_synth = pd.DataFrame(data_synth, columns=[c.attribute for c in self.network_])
         return data_synth
 
     def _sample_record(self):
         """samples a value column for column by conditioning for parents"""
         record = {}
         for col_idx, pair in enumerate(self.network_):
-            node = self.model_[pair.node]
+            node = self.model_[pair.attribute]
             node_cpt = node.cpt
             node_states = node.states
 
             if node.conditioning:
                 parent_values = [record[p] for p in node.conditioning]
                 node_probs = node_cpt[tuple(parent_values)]
+
             else:
                 node_probs = node_cpt.values
-            sampled_node_value = np.random.choice(node_states, p=node_probs)
+            # use random.choices over np.random.choice as np coerces, e.g. sample['nan', 1, 3.0] -> '1' (int to string)
+            sampled_node_value = random.choices(node_states, weights=node_probs, k=1)[0] # returns list
 
             record[node.name] = sampled_node_value
         return record
@@ -293,6 +308,8 @@ class PrivBayes(BaseDPSynthesizer):
         Relies on the L1 distance from a joint distribution to a joint distributions that minimizes mutual information.
         Refer to Lemma 5.2
         """
+        if columns_b is None:
+            return 0
         columns_a = utils._ensure_arg_is_list(columns_a)
         columns_b = utils._ensure_arg_is_list(columns_b)
 
@@ -310,6 +327,45 @@ class PrivBayes(BaseDPSynthesizer):
         l1_distance = 0.5 * np.sum(np.abs(prob_joint.values - prob_independent.values))
         return l1_distance
 
+    def save(self, path):
+        """
+        Save this synthesizer instance to the given path using pickle.
+
+        Parameters
+        ----------
+        path: str
+            Path where the synthesizer instance is saved.
+        """
+        # todo issue can't save if model is fitted - likely error within thomas
+        if hasattr(self, 'model_'):
+            pb = self.copy()
+            del pb.model_
+
+        with open(path, 'wb') as output:
+            dill.dump(pb, output)
+
+    @classmethod
+    def load(cls, path):
+        """Load a synthesizer instance from specified path.
+        Parameters
+        ----------
+        path: str
+            Path where the synthesizer instance is saved.
+
+        Returns
+        -------
+        synthesizer : class
+            Returns synthesizer instance.
+        """
+        with open(path, 'rb') as f:
+            pb = dill.load(f)
+
+        # recreate model_ attribute based on fitted network and cpt's
+        if hasattr(pb, 'cpt_'):
+            pb.model_ = BayesianNetwork.from_CPTs('PrivBayes', pb.cpt_.values())
+        return pb
+
+
 
 class PrivBayesFix(PrivBayes):
     """Extension to PrivBayes class to allow user to fix pre-sampled columns. Can be used to generate additional items
@@ -317,9 +373,9 @@ class PrivBayesFix(PrivBayes):
     """
 
     def __init__(self, epsilon=1.0, theta_usefulness=4, epsilon_split=0.3,
-                 score_function='R', network_init=None, verbose=True):
+                 score_function='R', network_init=None, n_cpus=None, verbose=True):
         super().__init__(epsilon=epsilon, theta_usefulness=theta_usefulness, epsilon_split=epsilon_split,
-                         score_function=score_function, network_init=network_init, verbose=verbose)
+                         score_function=score_function, network_init=network_init, n_cpus=n_cpus, verbose=verbose)
 
     def fit(self, data):
         super().fit(data)
@@ -349,7 +405,7 @@ class PrivBayesFix(PrivBayes):
             data_synth[i] = list(record.values())
 
         # numpy.array to pandas.DataFrame with original column ordering
-        data_synth = pd.DataFrame(data_synth, columns=[c.node for c in self.network_])
+        data_synth = pd.DataFrame(data_synth, columns=[c.attribute for c in self.network_])
         return data_synth
 
     def _sample_record(self, record_init):
@@ -359,7 +415,7 @@ class PrivBayesFix(PrivBayes):
 
         # sample remaining nodes after fixing for input X
         for col_idx, pair in enumerate(self.network_[len(record_init):]):
-            node = self.model_[pair.node]
+            node = self.model_[pair.attribute]
 
             # specify pre-sampled conditioning values
             node_cpt = node.cpt
@@ -370,7 +426,8 @@ class PrivBayesFix(PrivBayes):
                 node_probs = node_cpt[tuple(parent_values)]
             else:
                 node_probs = node_cpt.values
-            sampled_node_value = np.random.choice(node_states, p=node_probs)
+            # sampled_node_value = np.random.choice(node_states, p=node_probs)
+            sampled_node_value = random.choices(node_states, weights=node_probs)
             record[node.name] = sampled_node_value
         return record
 
@@ -390,17 +447,37 @@ class PrivBayesFix(PrivBayes):
         # prevent integer column indexing issues
         data.columns = data.columns.astype(str)
         # make all columns categorical.
-        data = data.astype(str)
+        data = data.astype('category')
         return data
 
+class PrivBayesNP(PrivBayes):
+    """Privbayes class with infinite-differential privacy, while still using epsilon value to limit the size of the network
+    """
+
+    def __init__(self, epsilon=1, theta_usefulness=4, epsilon_split=0.3,
+                 score_function='R', network_init=None, n_cpus=None, verbose=True):
+        self.epsilon1 = float(np.inf)
+        self.epsilon2 = epsilon
+        super().__init__(epsilon=self.epsilon1, theta_usefulness=theta_usefulness, epsilon_split=epsilon_split,
+                         score_function=score_function, network_init=network_init, n_cpus=n_cpus, verbose=verbose)
+
+
+    def _max_domain_size(self, data, node):
+        """Computes the maximum domain size a node can have to satisfy theta-usefulness"""
+        node_cardinality = utils.cardinality(data[node])
+        max_domain_size = self.n_records_fit_ * self.epsilon2 / \
+                          (2 * len(self.columns_) * self.theta_usefulness * node_cardinality)
+        return max_domain_size
+
+
 if __name__ == "__main__":
     data_path = '../../examples/data/original/adult.csv'
     data = pd.read_csv(data_path, engine='python')
-    columns = ['age', 'sex', 'education', 'workclass', 'income']
-    data = data.loc[:, columns]
+    # columns = ['age', 'sex', 'education', 'workclass', 'income']
+    # data = data.loc[:, columns]
     print(data.head())
 
-    pb = PrivBayes(epsilon=0.1)
+    pb = PrivBayes(epsilon=1, n_cpus=4)
     pb.fit(data)
     df_synth = pb.sample(1000)
     pb.score(data, df_synth, score_dict=True)
@@ -416,7 +493,7 @@ if __name__ == "__main__":
     # df_synth_init = pbinit.sample(1000)
 
     # fixing a network - specify init network to fix those variables when generating
-    pbfix = PrivBayesFix(epsilon=1)
+    pbfix = PrivBayesFix(epsilon=1, n_cpus=4)
     init_network = [APPair('age', None), APPair('education', ('age',))]
     # init_network = [APPair(node='age', parents=None),
     #                 APPair(node='education', parents=('age',)),
@@ -430,6 +507,7 @@ if __name__ == "__main__":
 
     """test scoring functions"""
     pair = pb.network_[3]
-    pb.mi_score(data, pair.node, pair.parents)
-    pb.mi_score_thomas(data, pair.node, pair.parents)
-    pb.r_score(data, pair.node, pair.parents)
\ No newline at end of file
+    pb.mi_score(data, pair.attribute, pair.parents)
+    pb.mi_score_thomas(data, pair.attribute, pair.parents)
+    pb.r_score(data, pair.attribute, pair.parents)
+
