@@ -1,6 +1,6 @@
 """Transformer for preprocessing input data and postprocessing output data"""
 
-from synthesis.transformers.generalization import GeneralizeContinuous, GeneralizeCategorical, GeneralizeSchematic
+from synthesis.transformers.generalization import *
 from synthesis.transformers.sequence import GeneralizeDateSequence
 from synthesis.transformers.deidentification import *
 
